from ewah.constants import EWAHConstants as EC
from ewah.hooks.base import EWAHBaseHook
from ewah.uploaders.base import EWAHBaseUploader

from airflow.operators.python import PythonOperator
from airflow.models import BaseOperator

from google.cloud.bigquery.table import TableReference
from google.cloud.bigquery import (
    SchemaField,
    Table,
    LoadJobConfig,
    SourceFormat,
    CopyJobConfig,
)

from copy import deepcopy
from time import sleep
from tempfile import TemporaryFile, TemporaryDirectory
from datetime import datetime, date, timedelta


class BigqueryOperator(BaseOperator):
    "Operate to execute SQL on BigQuery"

    def __init__(
        self, sql, bigquery_conn_id, project=None, params=None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.sql = sql
        self.conn_id = bigquery_conn_id
        self.project = project
        # an attribute called "params" of an operator leads to an airflow error
        # if that attribute is None due to default templating of "task.params"!
        self._params = params

    def execute(self, context):
        from ewah.hooks.bigquery import EWAHBigQueryHook

        hook = EWAHBigQueryHook(
            conn_id=self.conn_id,
            project_id=self.project,
        )
        hook.execute(self.sql, commit=True, params=self._params)
        hook.commit()
        hook.close()


class FakeDatasetRef:
    def __init__(self, dataset_id, project_id):
        self.dataset_id = dataset_id
        self.project = project_id


class EWAHBigQueryUploader(EWAHBaseUploader):

    _QUERY_SCHEMA_CHANGES_COLUMNS = """
        SELECT column_name
        FROM `{project_id}.{schema_name}.INFORMATION_SCHEMA.COLUMNS`
        WHERE table_name = '{table_name}'
    """
    _QUERY_SCHEMA_CHANGES_ADD_COLUMN = """
        ALTER TABLE `{project_id}.{schema_name}.{table_name}`
        ADD COLUMN `{column_name}` {column_type};
    """

    _QUERY_TABLE = "SELECT * FROM `{project_id}.{schema_name}.{table_name}`"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(EC.DWH_ENGINE_BIGQUERY, *args, **kwargs)
        # BigQuery project id aka database name may be set in connection
        self.database_name = self.database_name or self.dwh_hook.conn.project

    @classmethod
    def cleaner_callables(cls):
        def bigquery_data_adjustments(row):
            key_changes = []
            for key, value in row.items():
                # change date, datetime, timedelta types
                # reason: non-json-serializable types are not accepted
                if isinstance(value, (date, datetime)):
                    row[key] = value.isoformat()
                elif isinstance(value, timedelta):
                    row[key] = value.total_seconds()

                # prefix field names that start with a number because BigQuery
                # does not like those types of field names.
                # Sub-challenge: generally prefix all field names that are hex
                try:
                    int(key, 16)
                    # if that did not fail, it is a hex!
                    key_changes.append((key, "field_" + key))
                except ValueError:
                    # not a hex, does it start with a number still?
                    if key[:1].isdigit():
                        key_changes.append((key, "field_" + key))

            # execute key changes
            for old_key, new_key in key_changes:
                row[new_key] = row.pop(old_key)

            return row

        return [bigquery_data_adjustments]

    @classmethod
    def get_schema_tasks(
        cls,
        dag,
        dwh_engine,
        dwh_conn_id,
        target_schema_name,
        target_schema_suffix="_next",
        target_database_name=None,
        read_right_users=None,  # Only for PostgreSQL
        **additional_task_args,
    ):
        schema_full_path = ""
        if target_database_name:
            schema_full_path += target_database_name + "."
        schema_full_path += target_schema_name

        # Create functions that are subsequently used in a PythonOperator

        def kickoff_func(schema_full, schema_suffix, dwh_conn_id):
            # kickoff: create new dataset
            schema = schema_full + schema_suffix
            conn = EWAHBaseHook.get_hook_from_conn_id(dwh_conn_id).dbconn
            # delete dataset first if it already exists
            print("Deleting the dataset {0} if it already exists.".format(schema))
            conn.delete_dataset(schema, delete_contents=True, not_found_ok=True)
            print("Creating the dataset {0}.".format(schema_full))
            conn.create_dataset(schema)
            print("Done!")

        def final_func(schema_name, schema_suffix, dwh_conn_id):
            # final: move new data into the final dataset
            conn = EWAHBaseHook.get_hook_from_conn_id(dwh_conn_id).dbconn
            # get dataset objects
            try:  # create final dataset if not exists
                ds_final = conn.get_dataset(schema_name)
            except:
                print("Creating dataset {0}".format(schema_name))
                ds_final = conn.create_dataset(schema_name)
            ds_temp = conn.get_dataset(schema_name + schema_suffix)

            # copy all tables from temp dataset to final dataset
            new_tables = conn.list_tables(ds_temp)
            new_table_ids = [table.table_id for table in conn.list_tables(ds_temp)]
            old_table_ids = [table.table_id for table in conn.list_tables(ds_final)]
            copy_jobs = []
            for table in new_tables:
                print(
                    "Copying table {0} from temp to final dataset".format(
                        table.table_id
                    )
                )
                try:
                    old_table = conn.get_table(
                        table=TableReference(
                            dataset_ref=ds_final, table_id=table.table_id
                        )
                    )
                    conn.delete_table(old_table)
                except:
                    # ignore failure, fails if old table does not exist to begin with
                    pass
                finally:
                    final_table = ds_final.table(table.table_id)
                    copy_jobs.append(conn.copy_table(table, final_table))

            # delete tables that don't exist in temp dataset from final dataset
            for table_id in old_table_ids:
                if not table_id in new_table_ids:
                    print("Deleting table {0}".format(table_id))
                    conn.delete_table(
                        conn.get_table(
                            TableReference(dataset_ref=ds_final, table_id=table_id)
                        )
                    )

            # make sure all copy jobs succeeded
            while copy_jobs:
                sleep(0.1)
                job = copy_jobs.pop(0)
                job.result()
                assert job.state in ("RUNNING", "DONE")
                if job.state == "RUNNING":
                    copy_jobs.append(job)
                else:
                    print(
                        "Successfully copied {0}".format(
                            job.__dict__["_properties"]["configuration"]["copy"][
                                "destinationTable"
                            ]["tableId"]
                        )
                    )

            # delete temp dataset
            print("Deleting temp dataset.")
            conn.delete_dataset(ds_temp, delete_contents=True, not_found_ok=False)

            print("Done.")

        task_1_args = deepcopy(additional_task_args)
        task_2_args = deepcopy(additional_task_args)
        task_1_args.update(
            {
                "task_id": "kickoff",
                "dag": dag,
                "python_callable": kickoff_func,
                "op_kwargs": {
                    "schema_full": schema_full_path,
                    "schema_suffix": target_schema_suffix,
                    "dwh_conn_id": dwh_conn_id,
                },
            }
        )
        task_2_args.update(
            {
                "task_id": "final",
                "dag": dag,
                "python_callable": final_func,
                "op_kwargs": {
                    "schema_name": target_schema_name,
                    "schema_suffix": target_schema_suffix,
                    "dwh_conn_id": dwh_conn_id,
                },
            }
        )

        return (PythonOperator(**task_1_args), PythonOperator(**task_2_args))

    def rollback(self):
        raise Exception("Rollback attempted - BigQuery knows no transactions!")

    def close(self):
        pass  # nothing to do

    def commit(self):
        pass  # nothing to do

    def test_if_table_exists(self, table_name, schema_name, project_id=None):

        if project_id:
            dataset_path = "{0}.{1}".format(project_id, schema_name)
        else:
            dataset_path = schema_name
        if not self.test_if_dataset_exists(dataset_path):
            return False
        return 0 < len(
            self.dwh_hook.execute_and_return_result(
                sql="""
                        SELECT * FROM `{dataset}.__TABLES__`
                        WHERE table_id = '{table_name}'
                    """.format(
                    dataset=dataset_path, table_name=table_name
                )
            )
        )

    def test_if_dataset_exists(self, dataset_id):
        conn = self.dwh_hook.dbconn
        try:
            ds = conn.get_dataset(dataset_id)
        except:
            # fails if dataset does not exist
            return False
        return True

    def get_max_value_of_column(self, column_name):
        return self.dwh_hook.execute_and_return_result(
            sql="SELECT MAX(`{0}`) FROM `{1}.{2}.{3}`".format(
                column_name,
                self.database_name,
                self.schema_name + self.schema_suffix,
                self.table_name,
            ),
            return_dict=False,
        )[0][0]

    def drop_table_if_exists(self, table_name, schema_name, project_id=None):
        if self.test_if_table_exists(
            table_name=table_name,
            schema_name=schema_name,
            project_id=project_id,
        ):
            conn = self.dwh_hook.dbconn
            conn.delete_table(
                conn.get_table(
                    TableReference(
                        dataset_ref=FakeDatasetRef(
                            dataset_id=schema_name,
                            project_id=project_id or self.database_name,
                        ),
                        table_id=table_name,
                    )
                )
            )

    def copy_table(self):
        # Overwrite parent method for alternative approach
        # If target table was created by a previous run, drop it first
        self.drop_table_if_exists(
            self.table_name, self.schema_name + self.schema_suffix, self.database_name
        )
        # Create target table if source table exists
        if self.test_if_table_exists(
            table_name=self.table_name,
            schema_name=self.schema_name,
            project_id=self.database_name,
        ):
            self.log.info("Copying table into temporary dataset!")
            conn = self.dwh_hook.dbconn
            ds_old = conn.get_dataset(self.schema_name)
            ds_new = conn.get_dataset(self.schema_name + self.schema_suffix)
            table_old = conn.get_table(
                table=TableReference(dataset_ref=ds_old, table_id=self.table_name)
            )
            table_new = ds_new.table(self.table_name)
            copy_job = conn.copy_table(
                table_old,
                table_new,
                job_config=CopyJobConfig(write_disposition="WRITE_TRUNCATE"),
            )
            sleep(1)

            while True:
                copy_job.result()
                assert copy_job.state in (
                    "RUNNING",
                    "DONE",
                ), "Unexpected job state: {0}".format(job.state)
                if copy_job.state == "DONE":
                    self.log.info(
                        "Successfully copied {0}".format(
                            copy_job.__dict__["_properties"]["configuration"]["copy"][
                                "destinationTable"
                            ]["tableId"]
                        )
                    )
                    break
                # Wait 5s, try again
                sleep(5)

    def _create_or_update_table(
        self,
        data,
        table_name,
        schema_name,
        schema_suffix,
        columns_definition,
        load_strategy,
        upload_call_count,
        database_name=None,
        primary_key=None,
    ):
        project_id = database_name or self.database_name
        conn = self.dwh_hook.dbconn
        new_schema_name = schema_name + schema_suffix
        ds_new = conn.get_dataset(new_schema_name)

        # ensure all fields exist, even if null
        # otherwise, rarely-populated fields will cause data loading failure
        upload_data = []
        while data:
            datum = data.pop(0)
            upload_data.append(
                {field: datum.get(field) for field in columns_definition.keys()}
            )

        # create table if it does not yet exist / drop if it needs dropping
        table_exists = self.test_if_table_exists(
            table_name=table_name,
            schema_name=new_schema_name,
            project_id=project_id,
        )

        # autodetect only works for json, csv
        # Future: add clustering_fields kwargs
        # job_config = LoadJobConfig(source_format=SourceFormat.JSON, autodetect=True)
        job_config = LoadJobConfig(autodetect=True)

        if (load_strategy == EC.LS_INSERT_REPLACE and upload_call_count == 1) or (
            not table_exists
        ):
            if table_exists:
                # Drop table before re-creating it
                conn.delete_table(
                    conn.get_table(
                        TableReference(dataset_ref=ds_new, table_id=table_name)
                    )
                )
            # create it anew
            table_obj = Table(".".join([project_id, new_schema_name, table_name]))
            job = conn.load_table_from_json(
                json_rows=upload_data, destination=table_obj, job_config=job_config
            )
            try:
                job.result()
            except:
                self.log.info("Errors occured - job errors: {0}".format(job.errors))
                raise
            assert job.state == "DONE", "Invalid job state: {0}".format(job.state)
        else:
            # table already exists, use `merge` statement to load data via temp table

            tmp_table_name = table_name + "__tmp"
            if self.test_if_table_exists(
                table_name=tmp_table_name,
                schema_name=new_schema_name,
                project_id=project_id,
            ):
                # Delete temporary table if it exists from a previous run
                conn.delete_table(
                    conn.get_table(
                        TableReference(dataset_ref=ds_new, table_id=tmp_table_name)
                    )
                )

            # create a temp table with new data
            self.log.info("Uploading data into a temp table...")
            table_obj = Table(".".join([project_id, new_schema_name, tmp_table_name]))
            job = conn.load_table_from_json(
                json_rows=upload_data, destination=table_obj, job_config=job_config
            )
            try:
                job.result()
            except:
                self.log.info("Errors occured - job errors: {0}".format(job.errors))
                raise
            assert job.state == "DONE", "Invalid job state: {0}".format(job.state)

            # merge it into existing table
            if load_strategy == EC.LS_UPSERT:
                merge_condition = " AND ".join(
                    ["TARGET.`{0}` = SOURCE.`{0}`".format(pk) for pk in primary_key]
                )
            elif load_strategy in (EC.LS_INSERT_ADD, EC.LS_INSERT_REPLACE):
                # never matched
                merge_condition = "FALSE"
            else:
                raise Exception("Not Implemented!")

            when_clauses = """
                WHEN MATCHED THEN
                    UPDATE SET {set_columns}
                WHEN NOT MATCHED THEN
                    INSERT ({columns})
                    VALUES ({columns})
            """.format(
                set_columns=",".join(
                    [
                        "`{0}` = SOURCE.`{0}`".format(field)
                        for field in columns_definition.keys()
                        if not field in (primary_key or [])
                    ]
                ),
                columns=", ".join(
                    ["`{0}`".format(field) for field in columns_definition.keys()]
                ),
            )

            sql = """
                MERGE INTO {target_name} AS TARGET
                USING {source_name} AS SOURCE
                ON {merge_condition}
                {when_clauses}
            """.format(
                target_name="`{0}.{1}.{2}`".format(
                    project_id, new_schema_name, table_name
                ),
                source_name="`{0}.{1}.{2}`".format(
                    project_id, new_schema_name, tmp_table_name
                ),
                merge_condition=merge_condition,
                when_clauses=when_clauses,
            )
            self.dwh_hook.execute(sql=sql)

            self.log.info("Deleting temp table...")
            # delete temp table
            conn.delete_table(
                conn.get_table(
                    TableReference(dataset_ref=ds_new, table_id=tmp_table_name)
                )
            )



from ewah.constants import EWAHConstants as EC
from ewah.hooks.plentymarkets import EWAHPlentyMarketsHook
from ewah.operators.base import EWAHBaseOperator

from datetime import datetime, date


class EWAHPlentyMarketsOperator(EWAHBaseOperator):

    _NAMES = ["plentymarkets"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
        EC.ES_SUBSEQUENT: True,
    }

    _CONN_TYPE = EWAHPlentyMarketsHook.conn_type

    def __init__(self, resource=None, additional_api_call_params=None, *args, **kwargs):
        kwargs["primary_key"] = kwargs.get("primary_key", "id")
        resource = resource or kwargs.get("target_table_name")
        if kwargs["extract_strategy"] == EC.ES_SUBSEQUENT:
            kwargs["subsequent_field"] = kwargs.get("subsequent_field", "updatedAt")
            # currently, only the orders and accounts/contacts resource works with subsequent loading
            assert any(["orders" in resource, "accounts/contacts" in resource])
        if kwargs["extract_strategy"] == EC.ES_INCREMENTAL:
            # currently, only the orders resource works with incremental loading
            assert (
                EWAHPlentyMarketsHook.format_resource(resource)
                in EWAHPlentyMarketsHook._INCREMENTAL_FIELDS.keys()
            ), "{0} is not incrementally loadable!".format(resource)
        super().__init__(*args, **kwargs)

        assert isinstance(additional_api_call_params, (type(None), dict))
        self.resource = resource
        self.additional_api_call_params = additional_api_call_params

    def ewah_execute(self, context):
        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            data_from = self.get_max_value_of_column(self.subsequent_field)
            # This is likely to be a text field - need a date instead
            if not isinstance(data_from, date):
                if isinstance(data_from, datetime):
                    data_from = data_from.date()
                elif isinstance(data_from, str):
                    data_from = datetime.fromisoformat(data_from).date()
                else:
                    raise Exception(
                        "Data type of {0} is invalid!".format(self.subsequent_field)
                    )
        else:
            data_from = self.data_from
        for batch in self.source_hook.get_data_in_batches(
            resource=self.resource,
            data_from=data_from,
            data_until=self.data_until,
            additional_params=self.additional_api_call_params,
        ):
            self.upload_data(batch)



from ewah.hooks.base import EWAHBaseHook

from dbtmetabase.models.interface import MetabaseInterface, DbtInterface

import os
import yaml


class EWAHMetabaseHook(EWAHBaseHook):

    _ATTR_RELABEL = {
        "user": "login",
        "database": "schema",
        "sync_timeout": "port",
    }

    conn_name_attr = "ewah_metabase_conn_id"
    default_conn_name = "ewah_metabase_default"
    conn_type = "ewah_metabase"
    hook_name = "EWAH Metabase Connection"

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["extra"],
            "relabeling": {
                "login": "User",
                "password": "Password",
                "host": "Host / URL",
                "schema": "Database Name (in Metabase)",
                "port": "Sync timeout (leave empty for none)",
            },
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        from wtforms import StringField, BooleanField

        return {
            "extra__ewah_metabase__http_string": StringField(
                "Use http instead of https?",
                default="uri",
            )
        }

    def push_dbt_docs_to_metabase(
        self,
        dbt_project_path: str,
        dbt_database_name: str,
    ):
        models, aliases = DbtInterface(
            path=None,
            manifest_path=os.sep.join(
                (
                    dbt_project_path,
                    yaml.load(
                        open(os.sep.join((dbt_project_path, "dbt_project.yml")), "rb"),
                        Loader=yaml.Loader,
                    )["target-path"],
                    "manifest.json",
                )
            ),
            database=dbt_database_name,
            schema_excludes=None,
            includes=None,
            excludes=None,
        ).read_models(
            include_tags=True,
            docs_url=None,
        )

        metabase = MetabaseInterface(
            host=self.conn.host.replace("http://", "").replace("https://", ""),
            user=self.conn.user,
            password=self.conn.password,
            use_http=(self.conn.http_string or "").lower().startswith(("y", "j")),
            verify=True,
            database=self.conn.database,
            sync=True,
            sync_timeout=self.conn.sync_timeout or None,
        )

        metabase.prepare_metabase_client(models)

        self.log.info("Pushing docs to dbt now!")
        metabase.client.export_models(
            database=metabase.database,
            models=models,
            aliases=aliases,
        )



from datetime import datetime, date, timedelta


class EWAHConstants:
    "This class contains a number of constants for use throughout Ewah."

    # Available DWH Engines - use all small caps!
    DWH_ENGINE_POSTGRES = "postgresql"
    DWH_ENGINE_SNOWFLAKE = "snowflake"
    DWH_ENGINE_BIGQUERY = "bigquery"
    DWH_ENGINE_REDSHIFT = "redshift"
    DWH_ENGINE_S3 = "s3"
    DWH_ENGINE_GS = "googlesheets"
    DWH_ENGINES = [
        DWH_ENGINE_POSTGRES,
        DWH_ENGINE_SNOWFLAKE,
        DWH_ENGINE_BIGQUERY,
        # DWH_ENGINE_REDSHIFT, To Be Implemented
        # DWH_ENGINE_S3, To Be Implemented
        DWH_ENGINE_GS,
    ]

    DWH_ENGINE_SPELLING_VARIANTS = {
        "bq": DWH_ENGINE_BIGQUERY,
        "postgres": DWH_ENGINE_POSTGRES,
        "snowflake": DWH_ENGINE_SNOWFLAKE,
        "sheets": DWH_ENGINE_GS,
        "gsheets": DWH_ENGINE_GS,
        "google sheets": DWH_ENGINE_GS,
        DWH_ENGINE_POSTGRES: DWH_ENGINE_POSTGRES,
        DWH_ENGINE_SNOWFLAKE: DWH_ENGINE_SNOWFLAKE,
        DWH_ENGINE_GS: DWH_ENGINE_GS,
        DWH_ENGINE_BIGQUERY: DWH_ENGINE_BIGQUERY,
    }

    # DAG Strategies
    DS_ATOMIC = "atomic"
    DS_MIXED = "mixed"
    DS_IDEMPOTENT = "idempotent"

    # Available Extract Strategies
    ES_FULL_REFRESH = "full-refresh"  # load all available data
    ES_INCREMENTAL = "incremental"  # just load data pertaining to a certain time
    ES_SUBSEQUENT = "subsequent"  # just load newest data

    # Available Load Strategies
    LS_UPSERT = "upsert"  # update data based on a (composite) primary key
    LS_INSERT_ADD = "insert_add"  # just append data
    LS_INSERT_REPLACE = "insert_replace"  # replace data
    LS_INSERT_DELETE = "insert_delete"  # delete relevant data before writing new data

    # Default load strategy depending on extract strategy
    DEFAULT_LS_PER_ES = {
        ES_FULL_REFRESH: LS_INSERT_REPLACE,
        ES_SUBSEQUENT: LS_UPSERT,
        ES_INCREMENTAL: LS_UPSERT,
    }

    # EC.LS_FULLCREMENTAL = 'fullcremental'
    # fullcremental is a mix of full refresh and incremental
    # --> not an independent load strategy!

    # Query Building Constants
    QBC_FIELD_TYPE = "data_type"

    """If a field is of inconsistent data type or type does not exist
    in the mapping below, use inconsistent value specified here."""
    QBC_TYPE_MAPPING = {
        DWH_ENGINE_POSTGRES: {
            str: "text",
            int: "bigint",
            float: "numeric",
            dict: "jsonb",
            list: "jsonb",
            tuple: "jsonb",
            set: "jsonb",
            frozenset: "jsonb",
            bool: "boolean",
            datetime: "timestamp with time zone",
            date: "date",
            timedelta: "interval",
        },
        DWH_ENGINE_SNOWFLAKE: {
            str: "VARCHAR",
            int: "NUMBER",
            float: "NUMBER",
            dict: "OBJECT",
            list: "ARRAY",
            tuple: "ARRAY",
            set: "ARRAY",
            frozenset: "ARRAY",
            bool: "BOOLEAN",
            datetime: "TIMESTAMP_TZ",
            date: "DATE",
        },
        DWH_ENGINE_BIGQUERY: {
            str: "STRING",
            int: "INT64",
            float: "NUMERIC",
            dict: "STRUCT",
            list: "STRUCT",
            tuple: "STRUCT",
            set: "STRUCT",
            frozenset: "STRUCT",
            bool: "BOOL",
            datetime: "TIMESTAMP",
            date: "DATE",
            bytes: "BYTES",
        },
        # DWH_ENGINE_REDSHIFT: {},
        # DWH_ENGINE_S3: {},
        DWH_ENGINE_GS: {  # must have these two values evaluate to True as bool
            # EWAH tries to dump all non-mapped types, thus map even if map is never used!
            str: "x",
            int: "x",
            float: "x",
            dict: "x",
            list: "x",
            tuple: "x",
            set: "x",
            frozenset: "x",
            bool: "x",
            datetime: "timestamp with time x",
            date: "x",
        },
    }



from airflow import DAG

from ewah.constants import EWAHConstants as EC
from ewah.utils.airflow_utils import datetime_utcnow_with_tz
from ewah.operators.base import EWAHBaseOperator
from ewah.uploaders import get_uploader

from collections.abc import Iterable
from copy import deepcopy
from datetime import datetime, timedelta
from typing import Optional, Type, Callable, List, Tuple, Union

import re


def dag_factory_atomic(
    dag_name: str,
    dwh_engine: str,
    dwh_conn_id: str,
    start_date: datetime,
    el_operator: Type[EWAHBaseOperator],
    operator_config: dict,
    target_schema_name: str,
    target_schema_suffix: str = "_next",
    target_database_name: Optional[str] = None,
    default_args: Optional[dict] = None,
    schedule_interval: timedelta = timedelta(days=1),
    end_date: Optional[datetime] = None,
    read_right_users: Optional[Union[List[str], str]] = None,
    additional_dag_args: Optional[dict] = None,
    additional_task_args: Optional[dict] = None,
    logging_func: Optional[Callable] = None,
    dagrun_timeout_factor: Optional[float] = None,
    task_timeout_factor: Optional[float] = None,
    **kwargs
) -> Tuple[DAG]:
    def raise_exception(msg: str) -> None:
        """Add information to error message before raising."""
        raise Exception("DAG: {0} - Error: {1}".format(dag_name, msg))

    logging_func = logging_func or print

    if kwargs:
        logging_func("unused config: {0}".format(str(kwargs)))

    additional_dag_args = additional_dag_args or {}
    additional_task_args = additional_task_args or {}

    if not read_right_users is None:
        if isinstance(read_right_users, str):
            read_right_users = [u.strip() for u in read_right_users.split(",")]
        if not isinstance(read_right_users, Iterable):
            raise_exception("read_right_users must be an iterable or string!")

    # fake catchup = True: between start_date and end_date is only one schedule_interval
    # --> run the full refreshs every schedule_interval at the same time instead of
    # having a drift in execution time!
    if end_date:
        end_date = min(end_date, datetime_utcnow_with_tz())
    else:
        end_date = datetime_utcnow_with_tz()
    start_date += int((end_date - start_date) / schedule_interval) * schedule_interval
    if start_date == end_date:
        # if the division result is a precise integer, that implies a definite end_date
        # --> adjust to get exactly one schedule_interval delta between start_date and
        # end_date to have one last run available (that should have run before end_date)
        start_date -= schedule_interval
    else:
        # Airflow executes after data_interval_end - start_date has to be
        # between exactly 1 and below 2 time schedule_interval before end_date!
        # end_date - 2*schedule_interval < start_date <= end_date - schedule_interval
        start_date -= schedule_interval
        # Make sure only one execution every runs scheduled but manual triggers work!
        end_date = start_date + 2 * schedule_interval - timedelta(seconds=1)

    if dagrun_timeout_factor:
        _msg = "dagrun_timeout_factor must be a number between 0 and 1!"
        assert isinstance(dagrun_timeout_factor, (int, float)) and (
            0 < dagrun_timeout_factor <= 1
        ), _msg
        additional_dag_args["dagrun_timeout"] = additional_dag_args.get(
            "dagrun_timeout", dagrun_timeout_factor * schedule_interval
        )

    if task_timeout_factor:
        additional_task_args["execution_timeout"] = additional_task_args.get(
            "execution_timeout", task_timeout_factor * schedule_interval
        )

    dag = DAG(
        dag_name,
        catchup=True,  # See above
        default_args=default_args,
        max_active_runs=1,
        schedule_interval=schedule_interval,
        start_date=start_date,
        end_date=end_date,
        **additional_dag_args,
    )

    kickoff, final = get_uploader(dwh_engine).get_schema_tasks(
        dag=dag,
        dwh_engine=dwh_engine,
        dwh_conn_id=dwh_conn_id,
        target_schema_name=target_schema_name,
        target_schema_suffix=target_schema_suffix,
        target_database_name=target_database_name,
        read_right_users=read_right_users,
        **additional_task_args,
    )

    base_config = deepcopy(additional_task_args)
    base_config.update(operator_config.get("general_config", {}))
    with dag:
        for table in operator_config["tables"].keys():
            table_config = deepcopy(base_config)
            table_config.update(operator_config["tables"][table] or {})
            table_config.update(
                {
                    "task_id": "extract_load_" + re.sub(r"[^a-zA-Z0-9_]", "", table),
                    "dwh_engine": dwh_engine,
                    "dwh_conn_id": dwh_conn_id,
                    "extract_strategy": table_config.get(  # Default to full refresh
                        "extract_strategy", EC.ES_FULL_REFRESH
                    ),
                    "target_table_name": operator_config["tables"][table].get(
                        "target_table_name", table
                    ),
                    "target_schema_name": target_schema_name,
                    "target_schema_suffix": target_schema_suffix,
                    "target_database_name": target_database_name,
                }
            )
            # Atomic DAG only works with full refresh and subsequent strategies!
            assert table_config["extract_strategy"] in (
                EC.ES_FULL_REFRESH,
                EC.ES_SUBSEQUENT,
            )
            table_config["load_strategy"] = table_config.get(
                "load_strategy",
                EC.DEFAULT_LS_PER_ES[table_config["extract_strategy"]],
            )
            table_task = el_operator(**table_config)
            kickoff >> table_task >> final

    return (dag,)



from ewah.operators.base import EWAHBaseOperator
from ewah.utils.python_utils import is_iterable_not_string
from ewah.constants import EWAHConstants as EC

from ewah.hooks.base import EWAHBaseHook as BaseHook

from datetime import datetime, timedelta
from pytz import timezone

from requests.auth import HTTPBasicAuth
import requests
import json
import time
import copy


class EWAHShopifyOperator(EWAHBaseOperator):

    _NAMES = ["shopify"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: False,
        EC.ES_INCREMENTAL: True,
    }

    _acceptable_api_versions = [
        "2020-07",
    ]
    _current_api_version = "2020-07"
    _base_url = "https://{shop}.myshopify.com/admin/api/{version}/{object}.json"

    # Set all accepted objects with their potential filter fields with defaults
    #   Note: always expected fields updated_at_max, updated_at_min, limit, id
    #   the key _is_drop_and_replace indicates objects that are only able
    #   to load with a full refresh
    _default_timestamp_fields = ("updated_at_min", "updated_at_max")
    _default_datetime_format = "%Y-%m-%d %H:%M:%S%z"
    _accepted_objects = {
        "balance_transactions": {
            "_is_drop_and_replace": True,
            "_object_url": "shopify_payments/balance/transactions",
            "_name_in_request_data": "transactions",
            "since_id": None,
            "last_id": None,
            "test": None,
            "payout_id": None,
            "payout_status": None,
        },
        "customers": {
            "ids": None,
            "since_id": None,
        },
        "disputes": {
            "_is_drop_and_replace": True,
            "_object_url": "shopify_payments/disputes",
        },
        "events": {
            "_timestamp_fields": ("created_at_min", "created_at_max"),
            "since_id": None,
            "filter": None,
            "verb": None,
        },
        "orders": {
            "ids": None,
            "since_id": None,
            "status": "any",
            "financial_status": None,
            "fulfillment_status": None,
            "fields": None,
        },
        "payouts": {
            "_timestamp_fields": ("date_min", "date_max"),
            "_datetime_format": "%Y-%m-%d",
            "_object_url": "shopify_payments/payouts",
            "since_id": None,
            "last_id": None,
            "status": None,
        },
        "products": {
            "ids": None,
            "since_id": None,
            "title": None,
            "vendor": None,
            "handle": None,
            "product_type": None,
            "collection_id": None,
            "published_status": None,
        },
        "tender_transactions": {
            "_timestamp_fields": ("processed_at_min", "processed_at_max"),
            "since_id": None,
        },
    }

    def __init__(
        self,
        shop_id,
        shopify_object,
        auth_type,
        filter_fields={},
        api_version=None,
        get_transactions_with_orders=False,
        get_events_with_orders=False,
        get_inventory_data_with_product_variants=False,
        page_limit=250,  # API Call pagination limit
        *args,
        **kwargs
    ):

        if is_iterable_not_string(shop_id):
            raise Exception("Multiple shops in one DAG is deprecated!")

        if get_transactions_with_orders and not shopify_object == "orders":
            raise Exception("transactions can only be pulled for orders!")

        if get_events_with_orders and not shopify_object == "orders":
            raise Exception("events can only be pulled for orders!")

        if (
            get_inventory_data_with_product_variants
            and not shopify_object == "products"
        ):
            raise Exception("inventory data may only be pulled with products!")

        if not shopify_object in self._accepted_objects.keys():
            raise Exception(
                "{0} is not in the list of accepted objects!"
                + " accepted objects: {1}".format(
                    shopify_object, ", ".join(self._accepted_objects.keys())
                )
            )

        if self._accepted_objects[shopify_object].get("_is_drop_and_replace"):
            kwargs["extract_strategy"] = EC.ES_FULL_REFRESH

        if not auth_type in ["access_token", "basic_auth"]:
            raise Exception("auth_type must be access_token or basic_auth!")

        if not type(filter_fields) == dict:
            raise Exception("filter_fields must be a dictionary!")
        else:
            for key, value in filter_fields.items():
                if not key in self._accepted_objects["shopify_object"].keys():
                    raise Exception(
                        "invalid key {0} in filter fields!".format(
                            key,
                        )
                    )

        api_version = api_version or self._current_api_version
        if not api_version in self._acceptable_api_versions:
            raise Exception(
                "{0} is not a valid api version! valid versions: {1}".format(
                    api_version,
                    ", ".join(self._acceptable_api_versions),
                )
            )

        if not type(page_limit) == int or page_limit > 250 or page_limit < 1:
            raise Exception("Page limit must be a positive integer not exceeding 250!")

        kwargs["primary_key"] = kwargs.get("primary_key", "id")

        # source conn id is not required on operator call level! avoid error
        kwargs["source_conn_id"] = kwargs.get("source_conn_id", "__none__")

        super().__init__(*args, **kwargs)

        self.shop_id = shop_id
        self.shopify_object = shopify_object
        self.auth_type = auth_type
        self.filter_fields = filter_fields
        self.api_version = api_version
        self.page_limit = page_limit
        self.get_transactions_with_orders = get_transactions_with_orders
        self.get_events_with_orders = get_events_with_orders
        self.get_inventory_data_with_product_variants = (
            get_inventory_data_with_product_variants
        )

    def ewah_execute(self, context):
        # can supply a list of shops - need to run for all shops individually!
        def datetime_to_string(dt, format):
            # check if tz aware; set to utc if so
            if dt.tzinfo:
                dt = dt.astimezone(timezone("UTC"))
            else:
                dt = dt.replace(tzinfo=timezone("UTC"))
            # check if format_string contains timezone
            if "%z" in format:
                # add colon!
                dt_string = dt.strftime(format)
                return dt_string[:-2] + ":" + dt_string[-2:]
            else:
                return dt.strftime(format)

        object_metadata = self._accepted_objects[self.shopify_object]
        self.object_metadata = object_metadata
        params = {
            key: val
            for key, val in object_metadata.items()
            if not val is None and not key[:1] == "_"
        }
        params.update(self.filter_fields)
        params["limit"] = self.page_limit

        timestamp_fields = object_metadata.get(
            "_timestamp_fields",
            self._default_timestamp_fields,
        )
        timestamp_format_string = object_metadata.get(
            "_datetime_format",
            self._default_datetime_format,
        )
        if self.data_until:
            params[timestamp_fields[1]] = datetime_to_string(
                self.data_until,
                timestamp_format_string,
            )
        if self.data_from:
            params[timestamp_fields[0]] = (
                datetime_to_string(
                    self.data_from,
                    timestamp_format_string,
                ),
            )

        source_conn_id = self.source_conn.conn_id
        auth_type = self.auth_type
        if is_iterable_not_string(self.shop_id):
            # multiple shops to iterate - loop through!
            # deprecated feature - don't use!
            raise Exception("Multiple Shops in one DAG is deactivated!")
            self.log.info("iterating through multiple shops!")
            for shop_id in self.shop_id:
                # metadata: shop id
                self.log.info("getting data for: {0}".format(shop_id))
                if hasattr(self.shop_id, "get") and hasattr(
                    self.shop_id[shop_id], "get"
                ):
                    # dict, not list! check for conn details!
                    shop_dict = self.shop_id[shop_id]
                    sci = shop_dict.get("source_conn_id", source_conn_id)
                    at = shop_dict.get("auth_type", auth_type)
                else:
                    sci = source_conn_id
                    at = auth_type
                self._metadata.update({"shop_id": shop_id})
                self.execute_for_shop(context, shop_id, params, sci, at)
        else:
            self._metadata.update({"shop_id": self.shop_id})
            sci = self.source_conn.conn_id
            at = self.auth_type
            self.execute_for_shop(context, self.shop_id, params, sci, at)

    def execute_for_shop(
        self,
        context,
        shop_id,
        params,
        source_conn_id,
        auth_type,
    ):
        # Get data from shopify via REST API
        def add_get_transactions(data, shop, version, req_kwargs):
            # workaround to add transactions to orders
            self.log.info("Requesting transactions of orders...")
            base_url = "https://{shop}.myshopify.com/admin/api/{version}/orders/{id}/transactions.json"
            base_url = base_url.format(
                **{
                    "shop": shop,
                    "version": version,
                    "id": "{id}",
                }
            )

            for datum in data:
                id = datum["id"]
                # self.log.info('getting transactions for order {0}'.format(id))
                time.sleep(1)  # avoid hitting api call requested per second limit
                url = base_url.format(id=id)
                req = requests.get(url, **req_kwargs)
                if not req.status_code == 200:
                    self.log.info("response: " + str(req.status_code))
                    self.log.info("request text: " + req.text)
                    raise Exception("non-200 response!")
                transactions = json.loads(req.text).get("transactions", [])
                datum["transactions"] = transactions

            return data

        def add_get_inventoryitems(data, shop, version, req_kwargs):
            # workaround to get inventory item data (i.e. costs) for products
            self.log.info("Requesting inventory items of product variants...")
            base_url = (
                "https://{shop}.myshopify.com/admin/api/{version}/inventory_items.json"
            )
            url = base_url.format(
                shop=shop,
                version=version,
            )

            kwargs = copy.deepcopy(req_kwargs)

            for datum in data:
                ids = [v["inventory_item_id"] for v in datum.get("variants", [])]
                if ids:
                    kwargs["params"] = {"ids": copy.deepcopy(ids)}
                    time.sleep(1)  # avoid hitting api call requested limit
                    req = requests.get(url, **kwargs)
                    if not req.status_code == 200:
                        self.log.info("response: " + str(req.status_code))
                        self.log.info("request text: " + req.text)
                        raise Exception("non-200 response!")
                    inv_items = json.loads(req.text).get("inventory_items", [])
                    datum["inventory_items"] = inv_items

            return data

        def add_get_events(data, shop, version, req_kwargs):
            # workaround to add events of an order to orders
            self.log.info("Requesting events of orders...")
            base_url = "https://{shop}.myshopify.com/admin/api/{version}/orders/{id}/events.json"
            base_url = base_url.format(
                shop=shop,
                version=version,
                id="{id}",
            )

            for datum in data:
                id = datum["id"]
                time.sleep(1)
                url = base_url.format(id=id)
                req = requests.get(url, **req_kwargs)
                if not req.status_code == 200:
                    self.log.info("response: " + str(req.status_code))
                    self.log.info("request text: " + req.text)
                    raise Exception("non-200 response!")
                events = json.loads(req.text).get("events", [])
                datum["events"] = events

            return data

        url = self._base_url.format(
            **{
                "shop": shop_id,
                "version": self.api_version,
                "object": self.object_metadata.get(
                    "_object_url",
                    self.shopify_object,
                ),
            }
        )

        # get connection for the applicable shop
        conn = BaseHook.get_connection(source_conn_id)
        login = conn.login
        password = conn.password

        if auth_type == "access_token":
            headers = {
                "X-Shopify-Access-Token": password,
            }
            kwargs_init = {
                "headers": headers,
                "params": params,
            }
            kwargs_links = {"headers": headers}
        elif auth_type == "basic_auth":
            kwargs_init = {
                "params": params,
                "auth": HTTPBasicAuth(login, password),
            }
            kwargs_links = {"auth": HTTPBasicAuth(login, password)}
        else:
            raise Exception("Authentication type not accepted!")

        # get and upload data
        self.log.info(
            "Requesting data from REST API - url: {0}, params: {1}".format(
                url, str(params)
            )
        )
        req_kwargs = kwargs_init
        is_first = True
        while is_first or (r.status_code == 200 and url):
            r = requests.get(url, **req_kwargs)
            if is_first:
                is_first = False
                req_kwargs = kwargs_links
            data = json.loads(r.text or "{}").get(
                self.object_metadata.get(
                    "_name_in_request_data",
                    self.shopify_object,
                )
            )
            if self.get_transactions_with_orders:
                data = add_get_transactions(
                    data=data,
                    shop=shop_id,
                    version=self.api_version,
                    req_kwargs=kwargs_links,
                )
            if self.get_events_with_orders:
                data = add_get_events(
                    data=data,
                    shop=shop_id,
                    version=self.api_version,
                    req_kwargs=kwargs_links,
                )
            if self.get_inventory_data_with_product_variants:
                data = add_get_inventoryitems(
                    data=data,
                    shop=shop_id,
                    version=self.api_version,
                    req_kwargs=kwargs_links,
                )
            self.upload_data(data)
            self.log.info("Requesting next page of data...")
            if r.headers.get("Link") and r.headers["Link"][-9:] == 'el="next"':
                url = r.headers["Link"][1:-13]
            else:
                url = None

        if not r.status_code == 200:
            raise Exception(
                "Shopify request returned an error {1}: {0}".format(
                    r.text,
                    str(r.status_code),
                )
            )



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.base import EWAHBaseHook as BaseHook

from datetime import datetime, timedelta

from requests.auth import HTTPBasicAuth
import requests
import json
import time


class EWAHZendeskOperator(EWAHBaseOperator):

    _NAMES = ["zendesk"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: False,
        EC.ES_INCREMENTAL: True,
    }

    _base_url = "https://{support_url}.zendesk.com/{endpoint}"

    def __init__(self, support_url, resource, auth_type, *args, **kwargs):

        self._accepted_resources = {
            "tickets": {
                "function": self.execute_incremental_cursor_based,
                "datetime_field": "updated_at",
            },
            "ticket_audits": {
                "function": self.execute_incremental_cursor_based,
                "datetime_field": "created_at",
            },
            "ticket_metric_events": {
                "function": self.execute_incremental_time_based,
            },
            "users": {
                "function": self.execute_incremental_time_based,
            },
            "ticket_fields": {
                "function": self.get_custom_ticket_fields,
                "drop_and_replace": True,
            },
        }

        if not resource in self._accepted_resources.keys():
            raise Exception('resource "{0}" not supported!'.format(resource))

        if not auth_type in ["basic_auth"]:
            raise Exception("auth_type must be basic_auth!")

        kwargs["primary_key"] = kwargs.get("primary_key", "id")

        if self._accepted_resources[resource].get("drop_and_replace"):
            kwargs["extract_strategy"] = EC.ES_FULL_REFRESH

        # API data is delayed by about 60s according to official docs
        kwargs["wait_for_seconds"] = max(kwargs.get("wait_for_seconds", 0), 70)

        super().__init__(*args, **kwargs)

        self.support_url = support_url
        self.resource = resource
        self.auth_type = auth_type

    def ewah_execute(self, context):
        self.z_data_from = self.make_unix_datetime(self.data_from)
        self.z_data_until = self.make_unix_datetime(self.data_until)

        conn = self.source_conn
        if self.auth_type == "basic_auth":
            self.auth = HTTPBasicAuth(conn.login, conn.password)
        else:
            raise Exception("auth_type not implemented!")

        self._metadata.update({"support_url": self.support_url})

        # run correct execute function
        return self._accepted_resources[self.resource]["function"](context, self)

    def make_unix_datetime(self, dt):
        return str(int(time.mktime(dt.timetuple())))

    def get_custom_ticket_fields(_, context, self):
        url = self._base_url.format(
            support_url=self.support_url,
            endpoint="api/v2/ticket_fields.json",
        )
        req = requests.get(url, auth=self.auth)
        if not req.status_code == 200:
            raise Exception(
                "Error {1} when calling Zendesk API: {0}".format(
                    req.text,
                    str(req.status_code),
                )
            )
        data = json.loads(req.text)["ticket_fields"]
        self.upload_data(data)

    def execute_incremental_cursor_based(_, context, self):
        if not self.resource in ["tickets", "ticket_audits"]:
            raise Exception("cursor-based load not implemented!")

        if self.resource == "tickets":
            endpoint = "api/v2/incremental/tickets/cursor.json"
            params = {
                "start_time": self.z_data_from,
                "include": "metric_sets",
            }
            response_resource = "tickets"
        elif self.resource == "ticket_audits":
            endpoint = "api/v2/ticket_audits.json"
            params = {
                "start_time": self.z_data_from,
            }
            response_resource = "audits"

        datetime_field = self._accepted_resources[self.resource]
        datetime_field = datetime_field["datetime_field"]

        url = self._base_url.format(
            support_url=self.support_url,
            endpoint=endpoint,
        )

        first_call = True
        while first_call or req.status_code == 200:
            first_call = False

            self.log.info("Requesting a page of data...")
            req = requests.get(url, params=params, auth=self.auth)
            response = json.loads(req.text)
            if req.status_code == 200:
                data = response[response_resource]
                if data:
                    last_updated_at = datetime.strptime(
                        data[-1][datetime_field],
                        "%Y-%m-%dT%H:%M:%SZ",
                    )
                    self.upload_data(data)
                if (
                    response.get("end_of_stream")
                    or last_updated_at > self.load_data_until
                ):
                    break
                params = {"cursor": response.get("after_cursor")}

        if not req.status_code == 200:
            raise Exception(
                "Error {1} when calling Zendesk API: {0}".format(
                    req.text,
                    str(req.status_code),
                )
            )

    def execute_incremental_time_based(_, context, self):

        url = self._base_url.format(
            **{
                "support_url": self.support_url,
                "endpoint": "api/v2/incremental/{0}.json".format(self.resource),
            }
        )

        params = {
            "start_time": self.z_data_from,
        }

        # Time-based incremental exports, run the first, the paginate!
        self.log.info(
            "Calling Zendesk api.\nurl: {0}\nparams:{1}".format(
                url,
                str(params),
            )
        )
        r = requests.get(url, params=params, auth=self.auth)
        data = json.loads(r.text)
        while (
            r.status_code == 200
            and not data.get("end_of_stream")
            and data.get("end_time")
            and data.get("end_time") <= int(self.z_data_until)
        ):
            self.upload_data(data[self.resource])  # uploads previous request
            self.log.info("Requesting next page of data...")
            r = requests.get(data["next_page"], auth=self.auth)  # new request
            data = json.loads(r.text)

        if not r.status_code == 200:
            raise Exception(
                "Error {1} when calling Zendesk API: {0}".format(
                    r.text,
                    str(r.status_code),
                )
            )

        # upload final request
        self.upload_data(data[self.resource])



import os, sys
from ewah.operators.base import EWAHBaseOperator

# import all operators by walking through all files in this directory and
# importing all objects that are subclasses of EWAHBaseOperator
# adapted from:
#   https://stackoverflow.com/questions/6246458/import-all-classes-in-directory
relevant_files = [
    f[:-3]
    for f in os.listdir(os.path.dirname(os.path.abspath(__file__)))
    if f.endswith(".py") and not f in ["__init__.py"]
]

operator_list = {}

for py_file in relevant_files:
    mod = __import__(".".join([__name__, py_file]), fromlist=[py_file])
    classes = [getattr(mod, x) for x in dir(mod) if isinstance(getattr(mod, x), type)]
    for cls in classes:
        if issubclass(cls, EWAHBaseOperator) and not cls == EWAHBaseOperator:
            names = cls._NAMES
            if isinstance(names, str):
                names = [names]
            for name in names:
                operator_list.update({name: cls})



from airflow.utils.log.logging_mixin import LoggingMixin

from ewah.constants import EWAHConstants as EC

from typing import List, Dict, Optional, Any, Callable, Union

from copy import deepcopy
from hashlib import sha256

# Refactor me
from bson.json_util import dumps  # dumping mongob objects to string
from collections import OrderedDict
from decimal import Decimal

import json


class EWAHJSONEncoder(json.JSONEncoder):
    """Extension of the native json encoder to deal with additional datatypes and
    issues relating to (+/-) Inf and NaN numbers.
    """

    def default(self, obj):
        """Method is called if an object cannot be serialized.

        Ought to return a serializeable value for the object. Ought to raise an error
        if unable to do so.

        Implemented types:
            - Decimal -> float
            - bson.objectid.ObjectId -> string
        """

        if isinstance(obj, Decimal):
            return float(obj)
        # Let the base class default method raise the TypeError
        return super().default(obj)

    def iterencode(self, o, _one_shot=False):
        """Overwrite the iterencode method because this is where the float
        special cases are handled in the floatstr() function. Copy-pasted
        original code and then adapted it to change floatstr() behavior.

        This was necessary because the json module accepts (+/-) Infinity and NaN
        objects as floats, but PostgreSQL uses the tighter JSON standard and does not
        accept them in json. The json module offers no optional flag to deal with this
        issue natively. Thus, overwrite the iterencode method and remove Inf and NaN
        by returning null instead.
        """
        if self.check_circular:
            markers = {}
        else:
            markers = None
        if self.ensure_ascii:
            _encoder = json.encoder.encode_basestring_ascii
        else:
            _encoder = json.encoder.encode_basestring

        def floatstr(
            o,
            allow_nan=self.allow_nan,
            _repr=float.__repr__,
            _inf=float("inf"),
            _neginf=-float("inf"),
        ):
            # Check for specials.  Note that this type of test is processor
            # and/or platform-specific, so do tests which don't depend on the
            # internals.
            if not (o != o or o == _inf or o == _neginf):
                return _repr(o)
            if not allow_nan:
                raise ValueError(
                    "Out of range float values are not JSON compliant: " + repr(o)
                )
            return "null"

        return json.encoder._make_iterencode(
            markers,
            self.default,
            _encoder,
            self.indent,
            floatstr,
            self.key_separator,
            self.item_separator,
            self.sort_keys,
            self.skipkeys,
            _one_shot,
        )(o, 0)


class EWAHCleaner(LoggingMixin):
    """Default data cleaner class for EWAH.

    The Cleaner is a class that bundles methods used to clean data between
    fetching and loading. Minor transformations may be executed herein.
    In addition, the Cleaner may keep track of the data schema.

    Derive from this class to extend functionalities.
    """

    def __init__(
        self,
        default_row: Optional[Dict[str, Any]] = None,
        add_metadata: bool = False,
        exclude_columns: Optional[List[str]] = None,
        hash_columns: Optional[List[str]] = None,
        rename_columns: Optional[Dict[str, str]] = None,
        additional_callables: Optional[Union[List[Callable], Callable]] = None,
        json_encoder: type = EWAHJSONEncoder,
    ):
        super().__init__()

        cleaning_steps = []

        if exclude_columns:
            cleaning_steps.append(self._exclude_columns)
            self.exclude_columns = exclude_columns

        if rename_columns:
            cleaning_steps.append(self._rename_columns)
            self.rename_columns = rename_columns

        if hash_columns:
            cleaning_steps.append(self._hash_row)
            self.hash_columns = hash_columns

        if add_metadata:
            cleaning_steps.append(self._add_metadata)
            self.add_metadata = add_metadata

        if additional_callables:
            if callable(additional_callables):
                cleaning_steps.append(additional_callables)
            else:
                cleaning_steps += additional_callables

        # Clean values right at the end
        cleaning_steps.append(self.clean_values)

        self.cleaning_steps = cleaning_steps
        self.default_row = default_row or {}
        self.json_encoder = json_encoder

        if default_row:
            # initialize with defaults
            self.fields_definition = {
                field: type(value) for field, value in default_row.items()
            }
        else:
            self.fields_definition = {}

    def _exclude_columns(self, row):
        for column in self.exclude_columns:
            row.pop(column, None)
        return row

    def _add_metadata(self, row):
        row.update(self.metadata)
        return row

    def _rename_columns(self, row):
        for (old_name, new_name) in self.rename_columns.items():
            row[new_name] = row.pop(old_name, row.get(new_name))
        return row

    def _hash_row(self, row):
        for column in self.hash_columns:
            row[column] = self._hash_value(row.get(column))
        return row

    @staticmethod
    def _hash_value(value):
        # Overwrite function for any other desired hashing behavior
        if value is None:
            return None
        return sha256(str(value).encode()).hexdigest()

    def clean_rows(
        self, rows: List[Dict[str, Any]], metadata: Optional[Dict[str, Any]] = None
    ) -> List[Dict[str, Any]]:
        cleaned_rows = []
        self.log.info("Cleaning {0} rows of data!".format(str(len(rows))))
        if self.add_metadata:
            self.metadata = metadata or {}
        while rows:
            cleaned_rows.append(self.clean_row(rows.pop(0)))
        return cleaned_rows

    def clean_values(self, raw_row: dict):
        row = deepcopy(self.default_row)

        while raw_row:
            key, value = raw_row.popitem()
            if not value is None:
                if isinstance(value, str):
                    if value == "\0":
                        # This is a null value -> treat as None
                        value = row.get(key)  # Use default, if exists
                    else:
                        # Some database systems don't handle this character well
                        # Thus, remove it
                        value = value.replace("\x00", "")
                elif isinstance(value, (dict, OrderedDict, list)):
                    # Logic copy-pasted from legacy - TODO: Refactor this!
                    try:
                        value = json.dumps(value, cls=self.json_encoder)
                    except TypeError:
                        # try dumping with bson utility function
                        # Refactor this, PLEASE!
                        value = dumps(value)
                elif isinstance(value, Decimal):
                    value = float(value)
                row[key] = value

                # Set the fields_definition for the key
                value_type = type(value)
                current_type_set = self.fields_definition.get(key)
                if current_type_set:
                    if not current_type_set == value_type:
                        # TODO: make this flexible
                        # For now, default to text in case of conflict
                        self.fields_definition[key] = str
                        self.log.info(
                            "WARNING! Data types are inconsistent. "
                            "Affected: {0}".format(key)
                        )
                else:
                    self.fields_definition[key] = value_type

        return row

    def clean_row(self, row: dict):
        for step in self.cleaning_steps:
            row = step(row)
        return row

    def get_columns_definition(self, dwh_engine):
        columns_definition = {}
        for field, datatype in self.fields_definition.items():
            data_type = EC.QBC_TYPE_MAPPING[dwh_engine].get(datatype)
            if not data_type:
                raise Exception(
                    "Field '{field}' has an invalid data type '{data_type}'!".format(
                        field=field, data_type=str(datatype)
                    )
                )
            columns_definition[field] = {EC.QBC_FIELD_TYPE: data_type}
        return columns_definition
